import json
import os
import re
import time
from datetime import datetime
from functools import lru_cache
from typing import List

import junit_xml
from pyaspeller import YandexSpeller
from tqdm import tqdm
from wasabi import Printer

from kontur.checkgrammar import parse

ya_speller = YandexSpeller(lang="ru", ignore_urls=True)


def all_lower(arr: List[str]) -> List[str]:
    return [x.strip().lower() for x in arr]


class Error:
    def __init__(self, form: str, element: str, text: str, errors: List[str]):
        self.form = form
        self.element = element
        self.text = text
        self.errors = errors


class GrammarError:
    def __init__(self, word: str, variants: List[str]):
        self.word = word
        self.variants = variants

    def __str__(self):
        return f"{self.word} -> {self.variants}"


def has_cyrillic(text):
    return bool(re.search("[а-яА-Я]", text))


@lru_cache(maxsize=2048)
def checkYaSpeller(text: str, our_dict=tuple()) -> List[GrammarError]:
    # Подготовить текст. Спеллер плохо проверяет слова с цифрами
    result: List[GrammarError] = []
    if not has_cyrillic(text):
        return result

    check = ya_speller.spell(text)
    for res in check:
        if (
            res["s"]
            and res["word"].lower() not in our_dict
            and res["word"].lower() not in all_lower(res["s"])
        ):  # Есть варианты замены
            result.append(GrammarError(res["word"], res["s"]))

    return result


class GrammarCheck:
    def __init__(self):
        self._src = []
        self._dict = []
        self._result = None
        self._skip = ""

    def update_dict_from_file(self, path_to_dict: str):
        """
        Дополняет словарь исключений
            :param path_to_dict: путь до файла со словарем
        """
        full_path = os.path.abspath(path_to_dict)
        assert os.path.exists(full_path), f"Не найден словарь {full_path}"

        with open(path_to_dict, "r", encoding="utf-8") as f:
            our_dict = all_lower(f.readlines())

        self._dict += our_dict

    def update_dict_from_bsl(self, bsl_settings="./.bsl-language-server.json"):
        """
        Дополняет словарь исключений из настроек bsl
            :param bsl_settings: путь до настроек .bsl-language-server.json
        """
        full_path = os.path.abspath(bsl_settings)
        assert os.path.exists(
            full_path
        ), f"Не найдены настройки bsl language server {full_path}"

        with open(full_path, "r", encoding="utf-8") as json_file:
            data = json.load(json_file)

        user_words_to_ignore = (
            data.get("diagnostics", {})
            .get("parameters", {})
            .get("Typo", {})
            .get("userWordsToIgnore", "")
        )
        assert (
            user_words_to_ignore
        ), "Не найден параметр diagnostics.parameters.Typo.userWordsToIgnore"

        our_dict = [x.strip().lower() for x in user_words_to_ignore.split(",")]

        self._dict += our_dict

    def add_src(self, src: str):
        """
        Добавить каталог с исходниками для обработки
            :param src: путь до каталога исходников
        """
        full_path = os.path.abspath(src)
        assert os.path.exists(full_path), f"Не найден каталог исходников {full_path}"

        self._src.append(src)

    def add_skip_pattern(self, pattern: str):
        """
        Добавить паттерн для исключения некоторых форм

        :param pattern: паттерн в формате glob
        """
        assert pattern, "Паттерн не должен быть пустым"
        self._skip = f"[!{pattern}]"

    def run(self):
        """
        Запуск проверки орфографии
        """
        assert self._src, "Не указаны каталоги для проверки"

        result = {}
        our_dict = tuple(self._dict)

        start_time = datetime.now()
        for src in self._src:
            objects = parse.parseSrc(src, self._skip)

            for obj, elements in objects.items():
                if len(self._src) == 1:
                    key = obj
                else:
                    key = f"{src}->{obj}"
                for element, text in tqdm(elements.items(), desc=obj):
                    if len(text) <= 3:
                        continue

                    text_to_check = " ".join(
                        [re.sub(r"[0-9]+", "", x) for x in text.split(" ")]
                    )

                    errors = checkYaSpeller(text_to_check, our_dict)

                    if errors:
                        if obj not in result:
                            result[key] = []
                        result[key].append(Error(obj, element, text, errors))
        msg = Printer()
        msg.divider("СТАТИСТИКА", char="*")
        msg.info(f"Время {datetime.now() - start_time}")
        msg.info(f"Обнаружено ошибок {sum([len(x) for x in result.items()])}")
        msg.info(f"{checkYaSpeller.cache_info()}")
        msg.divider("=", char="*")

        self._result = result

    def dump_junit(self, path_to_xml: str):
        """
        Формирует отчет в формате JUnit
            :param path_to_xml: Путь до выгружаемого отчета
        """

        assert self._result is not None, "Проверка еще не была выполнена"

        full_path = os.path.abspath(path_to_xml)
        ts = []
        for obj, details in self._result.items():
            test_cases = []
            for data in details:
                test_case = junit_xml.TestCase(
                    data.element,
                    classname=obj,
                    stderr=f"Возможно опечатка в тексте: {data.text}",
                )

                for error in data.errors:
                    test_case.add_failure_info(error, failure_type="WARNING")
                test_cases.append(test_case)

            ts.append(junit_xml.TestSuite(obj, test_cases))

        with open(full_path, "w", encoding="utf-8") as f:
            junit_xml.TestSuite.to_file(f, ts, encoding="utf-8")

    def print(self):
        """
        Вывод на экран результатов проверки
        """
        assert self._result is not None, "Проверка еще не была выполнена"
        msg = Printer()
        for obj, details in self._result.items():
            msg.divider(obj)
            for data in details:
                msg.warn(data.element)
                for error in data.errors:
                    msg.fail(error)

    def dump_dict(self, path: str):
        """
        Выгрузка отчета в файл. Чтобы было проще готовить файл исключений
            :param path: путь для выгрузки отчета
        """
        full_path = os.path.abspath(path)
        result = set()
        for obj, details in self._result.items():
            for data in details:
                for error in data.errors:
                    result.add(error.word.lower())

        with open(full_path, "w", encoding="utf-8") as f:
            for s in result:
                f.write(s + "\n")
            timeout = 30
            while not os.path.exists(full_path):
                time.sleep(1)
                timeout -= 1
                if timeout == 0:
                    print("Не удалось сохранить файл")

    @property
    def has_errors(self) -> bool:
        """
        Были ошибки при проверке
            :return: bool
        """
        assert self._result is not None, "Проверка еще не была выполнена"

        return bool(self._result)



from kontur.checkgrammar.grammar import GrammarCheck, checkYaSpeller


class TestGetDictionary:
    def test_update_dict_from_file(self):
        check = GrammarCheck()
        check.update_dict_from_file("tests/fixture/dictionary/dict.txt")
        assert len(check._dict) == 4, "Некорректно разобран словарь из файла"

    def test_update_two_dict_from_file(self):
        check = GrammarCheck()
        check.update_dict_from_file("tests/fixture/dictionary/dict.txt")
        assert len(check._dict) == 4, "Некорректно разобран словарь из файла"

        check.update_dict_from_file("tests/fixture/dictionary/dict2.txt")
        assert len(check._dict) == 9, "Некорректно разобран словарь из файла"

    def test_update_two_dict_from_file_reverse(self):
        check = GrammarCheck()

        check.update_dict_from_file("tests/fixture/dictionary/dict2.txt")
        assert len(check._dict) == 5, "Некорректно разобран словарь из файла"

        check.update_dict_from_file("tests/fixture/dictionary/dict.txt")
        assert len(check._dict) == 9, "Некорректно разобран словарь из файла"

    def test_update_dict_from_bsl(self):
        check = GrammarCheck()
        check.update_dict_from_bsl("tests/fixture/dictionary/bsl-language-server.json")
        assert len(check._dict) == 5, "Некорректно разобран словарь из файла"


def test_add_src():
    check = GrammarCheck()
    check.add_src("tests/fixture/epf_mistakes/")
    assert len(check._src) == 1, "Не удалось добавить каталог исходников"

    check.add_src("tests/fixture/epf_right/")
    assert len(check._src) == 2, "Не удалось добавить каталог исходников"


class TestYaSpeller:
    def test_check_ya_speller_error(self):
        result = checkYaSpeller("Конртагент")
        assert result, "Не найдено ошибок"

    def test_check_ya_speller_no_error(self):
        result = checkYaSpeller("Контрагент")
        assert not result, "Найдены ошибки"

    def test_check_ya_speller_no_error2(self):
        result = checkYaSpeller("Контрагент в Диадоке")
        assert not result, "Найдены ошибки"

    def test_check_ya_speller_no_error_with_dict(self):
        result = checkYaSpeller("КоНрТаГеНт  ", tuple(["конртагент"]))

        assert not result, "Найдены ошибки"

    def test_check_ya_speller_no_error_if_in_suggest(self):
        result = checkYaSpeller("Картинка состояние отчета отправлен")
        assert not result, "Найдены ошибки"

    def test_check_ya_speller_non_cyrillic_error(self):
        result = checkYaSpeller("to be on not to be")
        assert not result, "Найдены ошибки"


class TestRunResult:
    def test_run(self, check_errors):

        assert check_errors._result
        assert "ОбработкаСОшибками.Форма" in check_errors._result

        elements = check_errors._result["ОбработкаСОшибками.Форма"]

        assert len(elements) == 5

    def test_has_error(self, check_errors):
        assert check_errors.has_errors

    def test_run_no_error(self, check_no_errors):
        assert not check_no_errors._result

    def test_has_no_error(self, check_no_errors):
        assert not check_no_errors.has_errors


class TestDumpResults:
    def test_dump_junit(self, check_errors, temp_xml):
        check_errors.dump_junit(temp_xml)
        must_be = [
            'failures="5"',
            "ГраппаСПолями.Заголовок",
            "Реквизит2.Заголовок",
        ]
        with open(temp_xml, "r", encoding="utf-8") as f:
            xml = f.read()
            for x in must_be:
                assert x in xml

    def test_dump_txt(self, check_errors, temp_txt):
        check_errors.dump_dict(temp_txt)
        with open(temp_txt, "r", encoding="utf-8") as f:
            output = [x.strip() for x in f.readlines()]

        assert "граппа" in output

    def test_print(self, check_errors, capsys):
        check_errors.print()
        captured = capsys.readouterr()
        must_be = [
            "ОбработкаСОшибками.Форма",
            "ГраппаСПолями.Заголовок",
            "Реквизит2.Заголовок",
        ]
        for x in must_be:
            assert x in captured.out

    def test_dump_junit_no_errors(self, check_no_errors, temp_xml):
        check_no_errors.dump_junit(temp_xml)
        with open(temp_xml, "r", encoding="utf-8") as f:
            xml = f.read()
            assert xml == '<?xml version="1.0" encoding="utf-8"?>\n<testsuites/>\n'

    def test_dump_txt_no_errors(self, check_no_errors, temp_txt):
        check_no_errors.dump_dict(temp_txt)
        with open(temp_txt, "r", encoding="utf-8") as f:
            output = [x.strip() for x in f.readlines()]

        assert not output

    def test_print_no_errors(self, check_no_errors, capsys):
        check_no_errors.print()
        captured = capsys.readouterr()
        assert "" in captured.out


import os

from kontur.checkgrammar import parse


class TestGetRUContent:
    def test_get_ru_content_none(self):
        result = parse.getRuContent(None)
        assert result is None, "None должен возвращать None"

    def test_get_ru_content_ru(self, ru_content):
        result = parse.getRuContent(ru_content)
        assert result == "Тест", "Должен получить секцию RU строки"

    def test_get_ru_content_no_ru(self, no_ru_content):
        result = parse.getRuContent(no_ru_content)
        assert result is None, "RU секции нет"


class TestParseElements:
    def test_none_child_items(self):
        result = parse.getChildItems(None)
        assert not result

    def test_no_child_items(self, empty):
        result = parse.getChildItems(empty)
        assert not result

    def test_child_items_title(self, button):

        result = parse.getChildItems(button)

        assert result

        assert "Тест.Заголовок" in result
        assert result["Тест.Заголовок"] == "Тестовая кнопка"

    def test_child_items_tooltip(self, button):

        result = parse.getChildItems(button)

        assert result

        assert "Тест.Подсказка" in result
        assert result["Тест.Подсказка"] == "Это тестовая подсказка"

    def test_child_items_extended_tooltip(self, button):

        result = parse.getChildItems(button)

        assert result

        assert "Тест.РасшПодсказка" in result
        assert result["Тест.РасшПодсказка"] == "Это тестовая расширенная подсказка"

    def test_skip_elements_without_title(self, page_no_title):
        result = parse.getChildItems(page_no_title)

        assert result

        check_list = [
            "ПолеБезЗаголовка.Заголовок",
            "ФлагБезЗаголовка.Заголовок",
            "ПереключательБезЗаголовка.Заголовок",
        ]

        for key in check_list:

            assert key not in result

    def test_skip_picture(self, picture):
        result = parse.getChildItems(picture)

        assert not result

    def test_parse_form_with_auto_commandbar(self):
        path = os.path.abspath(
            "tests/fixture/epf_mistakes/ОбработкаСОшибками/Forms/Форма/Ext/Form.xml"
        )
        result = parse.parseForm(path)

        assert result
        check_list = [
            "КнопкаКоманднойПанели.Заголовок",
            "ГруппаСДекорациями.Заголовок",
            "ГраппаСПолями.Заголовок",
            "ГруппаСКнопками.Заголовок",
            "Надпись.Заголовок",
            "Реквизит1.Заголовок",
            "Реквизит2.Заголовок",
            "Команда1.Заголовок",
            "Флаг.Заголовок",
            "Флаг.Подсказка",
            "Флаг.РасшПодсказка",
        ]
        for key in check_list:
            assert key in result

    def test_parse_form_without_auto_commandbar(self):
        path = os.path.abspath(
            "tests/fixture/epf_mistakes/ОбработкаСОшибками/Forms/Тест_Форма/Ext/Form.xml"
        )
        result = parse.parseForm(path)

        assert result
        check_list = ["Реквизит1.Заголовок"]
        for key in check_list:
            assert key in result

    def test_parse_formatted_string(self):
        form = parse.parseSrc("tests/fixture/epf_right/")
        text = form["ОбработкаБезОшибок.Форма"]["ФорматированнаяСтрока.Заголовок"]
        test = "1. google.com\n2. гугл.рф\n3. строковая константа"
        assert text == test


class TestParseFolders:
    def test_parse_src(self):
        path = os.path.abspath("tests/fixture/epf_mistakes/")
        result = parse.parseSrc(path)

        assert result
        assert "ОбработкаСОшибками.Форма" in result

    def test_parse_and_exclude_src(self):
        path = os.path.abspath("tests/fixture/epf_mistakes/")
        result = parse.parseSrc(path, "[!Тест_*]*")

        assert result
        assert "ОбработкаСОшибками.Форма" in result
        assert "ОбработкаСОшибками.Тест_Форма" not in result



import os

from click.testing import CliRunner

from kontur.checkgrammar.cli import cli


class TestSrcOption:
    def test_empty_src(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["-o", "output.txt"])

        assert result.exit_code == 1
        assert "Необходимо указать каталоги для проверки" in result.output

    def test_skip(self):
        runner = CliRunner()
        result = runner.invoke(
            cli, ["--skip", "Тест_*", "./tests/fixture/epf_mistakes/"]
        )

        assert result.exit_code == 1
        assert "Обнаружены ошибки" in result.output
        assert "ОбработкаСОшибками.Форма" in result.output
        assert "ОбработкаСОшибками.Тест_Форма" not in result.output

    def test_skip_short(self):
        runner = CliRunner()
        result = runner.invoke(
            cli, ["-skip", "Тест_*", "./tests/fixture/epf_mistakes/"]
        )

        assert result.exit_code == 1
        assert "Обнаружены ошибки" in result.output
        assert "ОбработкаСОшибками.Форма" in result.output
        assert "ОбработкаСОшибками.Тест_Форма" not in result.output

    def test_multi_src(self):
        runner = CliRunner()
        result = runner.invoke(
            cli, ["./tests/fixture/epf_mistakes/", "./tests/fixture/epf_right/"]
        )

        assert result.exit_code == 1
        assert "ОбработкаСОшибками.Форма" in result.output
        assert "ОбработкаБезОшибок.Форма" in result.output


class TestResultCheck:
    def test_error(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["./tests/fixture/epf_mistakes/"])

        assert result.exit_code == 1
        assert "Обнаружены ошибки" in result.output
        assert "ОбработкаСОшибками.Форма" in result.output

    def test_no_error(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["./tests/fixture/epf_right/"])

        assert result.exit_code == 0
        assert "Нет ошибок" in result.output


class TestWorkWithDictionary:
    def test_error_dict(self):
        runner = CliRunner()
        result = runner.invoke(
            cli,
            [
                "--dict",
                "./tests/fixture/dictionary/dict.txt",
                "./tests/fixture/epf_mistakes/",
            ],
        )

        assert result.exit_code == 1
        assert "палями" not in result.output

    def test_error_dict_short(self):
        runner = CliRunner()
        result = runner.invoke(
            cli,
            [
                "-d",
                "./tests/fixture/dictionary/dict.txt",
                "./tests/fixture/epf_mistakes/",
            ],
        )

        assert result.exit_code == 1
        assert "палями" not in result.output

    def test_error_bsl(self):
        runner = CliRunner()
        result = runner.invoke(
            cli,
            [
                "--bsl-settings",
                "./tests/fixture/dictionary/bsl-language-server.json",
                "./tests/fixture/epf_mistakes/",
            ],
        )

        assert result.exit_code == 1
        assert "палями" not in result.output

    def test_error_bsl_short(self):
        runner = CliRunner()
        result = runner.invoke(
            cli,
            [
                "-bsl",
                "./tests/fixture/dictionary/bsl-language-server.json",
                "./tests/fixture/epf_mistakes/",
            ],
        )

        assert result.exit_code == 1
        assert "палями" not in result.output


class TestOutputOption:
    def test_error_junit(self, temp_xml):
        runner = CliRunner()
        result = runner.invoke(
            cli, ["--junit", temp_xml, "./tests/fixture/epf_mistakes/"]
        )

        assert result.exit_code == 1
        assert os.path.exists(temp_xml)

    def test_error_junit_short(self, temp_xml):
        runner = CliRunner()
        result = runner.invoke(cli, ["-j", temp_xml, "./tests/fixture/epf_mistakes/"])

        assert result.exit_code == 1
        assert os.path.exists(temp_xml)

    def test_output_dict(self, temp_txt):
        runner = CliRunner()
        result = runner.invoke(
            cli, ["./tests/fixture/epf_mistakes/", "--output", temp_txt]
        )

        assert result.exit_code == 1
        assert os.path.exists(temp_txt)

    def test_output_dict_short(self, temp_txt):
        runner = CliRunner()
        result = runner.invoke(cli, ["./tests/fixture/epf_mistakes/", "-o", temp_txt])

        assert result.exit_code == 1
        assert os.path.exists(temp_txt)
